# !OSLib.Source.Makefile/inc
#  Include file for making lower-level OSLib projects

# tools
DEFMOD		= do DefMod
REDUCEAOF	= ReduceAOF
RM		= rm -nq
MD		= cdir
AS		= do objasm -throwback
TOLOWER		= ToLower
MAKE		= amu -desktop
LIBFILE		= libfile
ECHO		= @echo $@
CP		= copy

# flags
COPYFLAGS   = A~C~DFLN~P~Q~R~S~T~V

# name rules
HDRS  	 = $(addprefix Hdr., ${SWIMODULES} ${MODULES})
OBJASM 	 = $(addprefix s., ${SWIMODULES})
OBJASM32 = $(addprefix s32., ${SWIMODULES})
LIBS  	 = $(addprefix o., ${SWIMODULES})
LIBS32 	 = $(addprefix o32., ${SWIMODULES})
CHEADERS = $(addprefix h., ${SWIMODULES} ${MODULES})
CSTRONG  = $(addprefix help., ${SWIMODULES} ${MODULES})
OBJS  	 = $(addprefix o., ${CMODULES})
OBJS32 	 = $(addprefix o32., ${CMODULES})

#------------------------------------------------------------------------
# file locations
SOURCEDIR	= <OSLib$SourceDir>
BYTEWIDE	= ${SOURCEDIR}.ByteWide
HELPDATA	= ${SOURCEDIR}.HelpData
OSLIB		= ${SOURCEDIR}.o.OSLib
OSLIB32		= ${SOURCEDIR}.o.OSLib32

.PHONY:	all oslib cstrong hdrs cheaders objasm objasm32 clean distclean cleanlib
.PHONY:	cleanfiles cleanhelp release_deeplib release_widelib release_help

# final targets
all:	oslib
oslib:	libs libs32 objasm objasm32 hdrs cheaders
	${ECHO}

cstrong:    ${CSTRONG}
      	    ${ECHO}

hdrs: 	    ${HDRS}
      	    ${ECHO}

cheaders:   ${CHEADERS}
      	    ${ECHO}

objasm:     ${OBJASM}
      	    ${ECHO}

objasm32:   	${OBJASM32}
      	    	${ECHO}

clean: 	    	cleanfiles cleanlib cleanhelp
      	    	${ECHO}

distclean:	clean
		${RM} $(addprefix s_0., ${SWIMODULES} )
		${RM} $(addprefix s32_0., ${SWIMODULES} )
		${RM} $(addprefix ${MAKEFILEDIR}., h Hdr l o o32 s s32 s_0 s32_0 help )
      	    	@echo

cleanlib:
   	${RM} ${LIBS}
   	${RM} ${LIBS32}

cleanfiles:
   	${RM} ${HDRS}
   	${RM} ${CHEADERS}
   	${RM} ${OBJASM}
   	${RM} ${OBJASM32}
   	${RM} ${OBJS} ${OBJS32}

cleanhelp:
   		${RM} ${CSTRONG}
		${RM} -r HelpData


release:
ifndef RELEASEDIR
	error Release Dir undefined
endif
	-${CP} s s_0		${COPYFLAGS}
	-${CP} s32 s32_0	${COPYFLAGS}
   	${MD} ${RELEASEDIR}
   	${CP} Hdr ${RELEASEDIR}.Hdr	${COPYFLAGS}
   	${CP} h ${RELEASEDIR}.h		${COPYFLAGS}

#------------------------------------------------------------------------


.SUFFIXES: .o .o32 .h .h++ .swi .ch .Hdr .ah .help .s .s32 .asm

.asm.o:	    	;@${MD} o
		@Set OSLib$Output $@
      	    	${AS} -I<OSLib$Path> -from $< -to <OSLib$Output>
      	    	@IfThere $@ then stamp $@ else create $@

.asm.o32:   	;@${MD} o
		@Set OSLib$Output $@
		${AS} -I<OSLib$Path> -Throwback -apcs 3/32bit -from $< -to <OSLib$Output>
      	    	@IfThere $@ then stamp $@ else create $@

.swi.s:     	;@${MD} s
		@Set OSLib$Output $@
      	    	${DEFMOD} -q -s -byte_wide ${BYTEWIDE} < $< > <OSLib$Output>
      	    	@IfThere $@ then stamp $@ else create $@

.swi.s32:   	;@${MD} s32
		@Set OSLib$Output $@
      	    	${DEFMOD} -q -s -byte_wide ${BYTEWIDE} -32bit < $< > <OSLib$Output>
      	    	@IfThere $@ then stamp $@ else create $@

.swi.h:	    	;@${MD} h
		@Set OSLib$Output $@
      	    	@${TOLOWER} OSLib$Output
      	    	${DEFMOD} -q -h < $< > <OSLib$Output>
      	    	@IfThere $@ then stamp $@ else create $@

.swi.Hdr:   	;@${MD} Hdr
		${DEFMOD} -q -hdr < $< > $@

.swi.h++:   	;${DEFMOD} -q -h++ < $< > $@

.swi.ch:    	;${DEFMOD} -q -chelp < $< > $@

.swi.ah:    	;${DEFMOD} -q -asmhelp < $< > $@

.swi.o:		;@${MD} ${SOURCEDIR}.o
		@${MD} l
		@${MD} o
      	    	@IfThere ${OSLIB} then else ${LIBFILE} -c -o ${OSLIB}
      	    	${DEFMOD} -q -l -o l.$* -byte_wide ${BYTEWIDE} < $<
      	    	@${REDUCEAOF} -via ViaFile
            	#@-${LIBFILE} -d ${OSLIB} l.$*.*
            	${LIBFILE} -i -o ${OSLIB} -v ViaFile
	    	@create $@
	    	@delete ViaFile
            	@Wipe l.$* ~C~FR~V
   	    	@echo

.swi.o32:	;@${MD} ${SOURCEDIR}.o
		@${MD} l
		@${MD} o32
      	    	@IfThere ${OSLIB32} then else ${LIBFILE} -c -o ${OSLIB32}
      	    	${DEFMOD} -q -l -o l.$* -byte_wide ${BYTEWIDE} -32bit < $<
	    	@${REDUCEAOF} -via ViaFile
            	#@-${LIBFILE} -d ${OSLIB32} l.$*.*
            	${LIBFILE} -i -o ${OSLIB32} -v ViaFile
            	@create $@
	    	@delete ViaFile
            	@Wipe l.$* ~C~FR~V
      	    	@echo

.swi.help:  	;@${MD} ${HELPDATA}
		@${MD} help
		${ECHO}
		${DEFMOD} -q -cstrong -o ${HELPDATA} < $<
            	@IfThere $@ then stamp $@ else Create $@
